@@ -44,11 +44,13 @@
 #include "apr.h"
 #include "apr_strings.h"
 #include "apr_lib.h"
+#include "apr_version.h"
+#if APR_MAJOR_VERSION < 2
+#include "apu_version.h"
+#endif
 #define APR_WANT_STRFUNC
 #include "apr_want.h"
 
-#define CORE_PRIVATE
-
 #include "httpd.h"
 #include "http_config.h"
 #include "http_core.h"
@@ -59,6 +61,10 @@
 #include "http_request.h"
 #include "util_script.h"
 #include "ap_mpm.h"
+#include "mpm_common.h"
+#include "ap_provider.h"
+#include <stdio.h>
+#include <stdlib.h>
 
 typedef struct
 {
@@ -73,6 +79,11 @@ typedef struct
 
 module AP_MODULE_DECLARE_DATA info_module;
 
+/* current file name when doing -DDUMP_CONFIG */
+static const char *dump_config_fn_info;
+/* file handle when doing -DDUMP_CONFIG */
+static apr_file_t *out = NULL;
+
 static void *create_info_config(apr_pool_t * p, server_rec * s)
 {
     info_svr_conf *conf =
@@ -98,34 +109,74 @@ static void put_int_flush_right(request_rec * r, int i, int field)
 {
     if (field > 1 || i > 9)
         put_int_flush_right(r, i / 10, field - 1);
-    if (i)
-        ap_rputc('0' + i % 10, r);
-    else
-        ap_rputs("&nbsp;", r);
+    if (i) {
+        if (r)
+            ap_rputc('0' + i % 10, r);
+        else
+            apr_file_putc((char)('0' + i % 10), out);
+    }
+    else {
+        if (r)
+            ap_rputs("&nbsp;", r);
+        else
+            apr_file_printf(out, " ");
+    }
 }
 
+static void set_fn_info(request_rec *r, const char *name)
+{
+    if (r)
+        ap_set_module_config(r->request_config, &info_module, (void *)name);
+    else
+        dump_config_fn_info = name;
+}
+
+static const char *get_fn_info(request_rec *r)
+{
+    if (r)
+        return ap_get_module_config(r->request_config, &info_module);
+    else
+        return dump_config_fn_info;
+}
+
+
 static void mod_info_indent(request_rec * r, int nest,
                             const char *thisfn, int linenum)
 {
     int i;
-    const char *prevfn =
-        ap_get_module_config(r->request_config, &info_module);
+    const char *prevfn = get_fn_info(r);
     if (thisfn == NULL)
         thisfn = "*UNKNOWN*";
     if (prevfn == NULL || 0 != strcmp(prevfn, thisfn)) {
-        thisfn = ap_escape_html(r->pool, thisfn);
-        ap_rprintf(r, "<dd><tt><strong>In file: %s</strong></tt></dd>\n",
+        if (r) {
+            thisfn = ap_escape_html(r->pool, thisfn);
+            ap_rprintf(r, "<dd><tt><strong>In file: %s</strong></tt></dd>\n",
                    thisfn);
-        ap_set_module_config(r->request_config, &info_module,
-                             (void *) thisfn);
+        }
+        else {
+            apr_file_printf(out, "# In file: %s\n", thisfn);
+        }
+        set_fn_info(r, thisfn);
     }
 
-    ap_rputs("<dd><tt>", r);
-    put_int_flush_right(r, linenum > 0 ? linenum : 0, 4);
-    ap_rputs(":&nbsp;", r);
+    if (r) {
+        ap_rputs("<dd><tt>", r);
+        put_int_flush_right(r, linenum > 0 ? linenum : 0, 4);
+        ap_rputs(":&nbsp;", r);
+    }
+    else if (linenum > 0) {
+        for (i = 1; i <= nest; ++i)
+            apr_file_printf(out, "  ");
+        apr_file_putc('#', out);
+        put_int_flush_right(r, linenum, 4);
+        apr_file_printf(out, ":\n");
+    }
 
     for (i = 1; i <= nest; ++i) {
-        ap_rputs("&nbsp;&nbsp;", r);
+        if (r)
+            ap_rputs("&nbsp;&nbsp;", r);
+        else
+            apr_file_printf(out, "  ");
     }
 }
 
@@ -133,18 +184,24 @@ static void mod_info_show_cmd(request_rec * r, const ap_directive_t * dir,
                               int nest)
 {
     mod_info_indent(r, nest, dir->filename, dir->line_num);
-    ap_rprintf(r, "%s <i>%s</i></tt></dd>\n",
-               ap_escape_html(r->pool, dir->directive),
-               ap_escape_html(r->pool, dir->args));
+    if (r)
+        ap_rprintf(r, "%s <i>%s</i></tt></dd>\n",
+                   ap_escape_html(r->pool, dir->directive),
+                   ap_escape_html(r->pool, dir->args));
+    else
+        apr_file_printf(out, "%s %s\n", dir->directive, dir->args);
 }
 
 static void mod_info_show_open(request_rec * r, const ap_directive_t * dir,
                                int nest)
 {
     mod_info_indent(r, nest, dir->filename, dir->line_num);
-    ap_rprintf(r, "%s %s</tt></dd>\n",
-               ap_escape_html(r->pool, dir->directive),
-               ap_escape_html(r->pool, dir->args));
+    if (r)
+        ap_rprintf(r, "%s %s</tt></dd>\n",
+                   ap_escape_html(r->pool, dir->directive),
+                   ap_escape_html(r->pool, dir->args));
+    else
+        apr_file_printf(out, "%s %s\n", dir->directive, dir->args);
 }
 
 static void mod_info_show_close(request_rec * r, const ap_directive_t * dir,
@@ -153,11 +210,17 @@ static void mod_info_show_close(request_rec * r, const ap_directive_t * dir,
     const char *dirname = dir->directive;
     mod_info_indent(r, nest, dir->filename, 0);
     if (*dirname == '<') {
-        ap_rprintf(r, "&lt;/%s&gt;</tt></dd>",
-                   ap_escape_html(r->pool, dirname + 1));
+        if (r)
+            ap_rprintf(r, "&lt;/%s&gt;</tt></dd>",
+                       ap_escape_html(r->pool, dirname + 1));
+        else
+            apr_file_printf(out, "</%s>\n", dirname + 1);
     }
     else {
-        ap_rprintf(r, "/%s</tt></dd>", ap_escape_html(r->pool, dirname));
+        if (r)
+            ap_rprintf(r, "/%s</tt></dd>", ap_escape_html(r->pool, dirname));
+        else
+            apr_file_printf(out, "/%s\n", dirname);
     }
 }
 
@@ -187,7 +250,7 @@ static int mod_info_module_cmds(request_rec * r, const command_rec * cmds,
     int shown = from;
     ap_directive_t *dir;
     if (level == 0)
-        ap_set_module_config(r->request_config, &info_module, NULL);
+        set_fn_info(r, NULL);
     for (dir = node; dir; dir = dir->next) {
         if (dir->first_child != NULL) {
             if (level < mod_info_module_cmds(r, cmds, dir->first_child,
@@ -234,20 +297,26 @@ typedef struct
     hook_get_t get;
 } hook_lookup_t;
 
-static hook_lookup_t startup_hooks[] = {
+static const hook_lookup_t startup_hooks[] = {
     {"Pre-Config", ap_hook_get_pre_config},
+    {"Check Configuration", ap_hook_get_check_config},
     {"Test Configuration", ap_hook_get_test_config},
     {"Post Configuration", ap_hook_get_post_config},
     {"Open Logs", ap_hook_get_open_logs},
+    {"Pre-MPM", ap_hook_get_pre_mpm},
+    {"MPM", ap_hook_get_mpm},
+    {"Drop Privileges", ap_hook_get_drop_privileges},
+    {"Retrieve Optional Functions", ap_hook_get_optional_fn_retrieve},
     {"Child Init", ap_hook_get_child_init},
     {NULL},
 };
 
-static hook_lookup_t request_hooks[] = {
+static const hook_lookup_t request_hooks[] = {
     {"Pre-Connection", ap_hook_get_pre_connection},
     {"Create Connection", ap_hook_get_create_connection},
     {"Process Connection", ap_hook_get_process_connection},
     {"Create Request", ap_hook_get_create_request},
+    {"Pre-Read Request", ap_hook_get_pre_read_request},
     {"Post-Read Request", ap_hook_get_post_read_request},
     {"Header Parse", ap_hook_get_header_parser},
     {"HTTP Scheme", ap_hook_get_http_scheme},
@@ -255,15 +324,34 @@ static hook_lookup_t request_hooks[] = {
     {"Quick Handler", ap_hook_get_quick_handler},
     {"Translate Name", ap_hook_get_translate_name},
     {"Map to Storage", ap_hook_get_map_to_storage},
-    {"Check Access", ap_hook_get_access_checker},
+    {"Check Access", ap_hook_get_access_checker_ex},
+    {"Check Access (legacy)", ap_hook_get_access_checker},
     {"Verify User ID", ap_hook_get_check_user_id},
+    {"Note Authentication Failure", ap_hook_get_note_auth_failure},
     {"Verify User Access", ap_hook_get_auth_checker},
     {"Check Type", ap_hook_get_type_checker},
     {"Fixups", ap_hook_get_fixups},
     {"Insert Filters", ap_hook_get_insert_filter},
     {"Content Handlers", ap_hook_get_handler},
-    {"Logging", ap_hook_get_log_transaction},
+    {"Transaction Logging", ap_hook_get_log_transaction},
     {"Insert Errors", ap_hook_get_insert_error_filter},
+    {"Generate Log ID", ap_hook_get_generate_log_id},
+    {NULL},
+};
+
+static const hook_lookup_t other_hooks[] = {
+    {"Monitor", ap_hook_get_monitor},
+    {"Child Status", ap_hook_get_child_status},
+    {"End Generation", ap_hook_get_end_generation},
+    {"Error Logging", ap_hook_get_error_log},
+    {"Query MPM Attributes", ap_hook_get_mpm_query},
+    {"Query MPM Name", ap_hook_get_mpm_get_name},
+    {"Register Timed Callback", ap_hook_get_mpm_register_timed_callback},
+    {"Extend Expression Parser", ap_hook_get_expr_lookup},
+    {"Set Management Items", ap_hook_get_get_mgmt_items},
+#if AP_ENABLE_EXCEPTION_HOOK
+    {"Handle Fatal Exceptions", ap_hook_get_fatal_exception},
+#endif
     {NULL},
 };
 
@@ -290,7 +378,7 @@ static int module_find_hook(module * modp, hook_get_t hook_get)
 
 static void module_participate(request_rec * r,
                                module * modp,
-                               hook_lookup_t * lookup, int *comma)
+                               const hook_lookup_t *lookup, int *comma)
 {
     if (module_find_hook(modp, lookup->get)) {
         if (*comma) {
@@ -351,20 +439,35 @@ static int show_server_settings(request_rec * r)
                "<dt><strong>Server Built:</strong> "
                "<font size=\"+1\"><tt>%s</tt></font></dt>\n",
                ap_get_server_built());
+    ap_rprintf(r,
+               "<dt><strong>Server loaded APR Version:</strong> "
+               "<tt>%s</tt></dt>\n", apr_version_string());
+    ap_rprintf(r,
+               "<dt><strong>Compiled with APR Version:</strong> "
+               "<tt>%s</tt></dt>\n", APR_VERSION_STRING);
+#if APR_MAJOR_VERSION < 2
+    ap_rprintf(r,
+               "<dt><strong>Server loaded APU Version:</strong> "
+               "<tt>%s</tt></dt>\n", apu_version_string());
+    ap_rprintf(r,
+               "<dt><strong>Compiled with APU Version:</strong> "
+               "<tt>%s</tt></dt>\n", APU_VERSION_STRING);
+#endif
     ap_rprintf(r,
                "<dt><strong>Module Magic Number:</strong> "
                "<tt>%d:%d</tt></dt>\n", MODULE_MAGIC_NUMBER_MAJOR,
                MODULE_MAGIC_NUMBER_MINOR);
     ap_rprintf(r,
                "<dt><strong>Hostname/port:</strong> "
-               "<tt>%s:%u</tt></dt>\n", ap_get_server_name(r),
+               "<tt>%s:%u</tt></dt>\n",
+               ap_escape_html(r->pool, ap_get_server_name(r)),
                ap_get_server_port(r));
     ap_rprintf(r,
                "<dt><strong>Timeouts:</strong> "
                "<tt>connection: %d &nbsp;&nbsp; "
                "keep-alive: %d</tt></dt>",
                (int) (apr_time_sec(serv->timeout)),
-               (int) (apr_time_sec(serv->timeout)));
+               (int) (apr_time_sec(serv->keep_alive_timeout)));
     ap_mpm_query(AP_MPMQ_MAX_DAEMON_USED, &max_daemons);
     ap_mpm_query(AP_MPMQ_IS_THREADED, &threaded);
     ap_mpm_query(AP_MPMQ_IS_FORKED, &forked);
@@ -403,10 +506,6 @@ static int show_server_settings(request_rec * r)
     ap_rputs(" -D OS=\"" OS "\"\n", r);
 #endif
 
-#ifdef APACHE_MPM_DIR
-    ap_rputs(" -D APACHE_MPM_DIR=\"" APACHE_MPM_DIR "\"\n", r);
-#endif
-
 #ifdef HAVE_SHMGET
     ap_rputs(" -D HAVE_SHMGET\n", r);
 #endif
@@ -478,7 +577,7 @@ static int show_server_settings(request_rec * r)
 #ifdef BUFFERED_LOGS
     ap_rputs(" -D BUFFERED_LOGS\n", r);
 #ifdef PIPE_BUF
-    ap_rputs(" -D PIPE_BUF=%ld\n", (long) PIPE_BUF, r);
+    ap_rprintf(r, " -D PIPE_BUF=%ld\n", (long) PIPE_BUF);
 #endif
 #endif
 
@@ -490,10 +589,6 @@ static int show_server_settings(request_rec * r)
     ap_rputs(" -D NEED_HASHBANG_EMUL\n", r);
 #endif
 
-#ifdef SHARED_CORE
-    ap_rputs(" -D SHARED_CORE\n", r);
-#endif
-
 /* This list displays the compiled in default paths: */
 #ifdef HTTPD_ROOT
     ap_rputs(" -D HTTPD_ROOT=\"" HTTPD_ROOT "\"\n", r);
@@ -503,10 +598,6 @@ static int show_server_settings(request_rec * r)
     ap_rputs(" -D SUEXEC_BIN=\"" SUEXEC_BIN "\"\n", r);
 #endif
 
-#if defined(SHARED_CORE) && defined(SHARED_CORE_DIR)
-    ap_rputs(" -D SHARED_CORE_DIR=\"" SHARED_CORE_DIR "\"\n", r);
-#endif
-
 #ifdef DEFAULT_PIDLOG
     ap_rputs(" -D DEFAULT_PIDLOG=\"" DEFAULT_PIDLOG "\"\n", r);
 #endif
@@ -515,10 +606,6 @@ static int show_server_settings(request_rec * r)
     ap_rputs(" -D DEFAULT_SCOREBOARD=\"" DEFAULT_SCOREBOARD "\"\n", r);
 #endif
 
-#ifdef DEFAULT_LOCKFILE
-    ap_rputs(" -D DEFAULT_LOCKFILE=\"" DEFAULT_LOCKFILE "\"\n", r);
-#endif
-
 #ifdef DEFAULT_ERRORLOG
     ap_rputs(" -D DEFAULT_ERRORLOG=\"" DEFAULT_ERRORLOG "\"\n", r);
 #endif
@@ -587,25 +674,108 @@ static int show_active_hooks(request_rec * r)
         ap_rputs("\n  </tt>\n</dt>\n", r);
     }
 
+    ap_rputs
+        ("</dl>\n<hr />\n<h2><a name=\"other_hooks\">Other Hooks</a></h2>\n<dl>",
+         r);
+
+    for (i = 0; other_hooks[i].name; i++) {
+        ap_rprintf(r, "<dt><strong>%s:</strong>\n <br /><tt>\n",
+                   other_hooks[i].name);
+        dump_a_hook(r, other_hooks[i].get);
+        ap_rputs("\n  </tt>\n</dt>\n", r);
+    }
+
     ap_rputs("</dl>\n<hr />\n", r);
 
     return 0;
 }
 
+static int cmp_provider_groups(const void *a_, const void *b_)
+{
+    const ap_list_provider_groups_t *a = a_, *b = b_;
+    int ret = strcmp(a->provider_group, b->provider_group);
+    if (!ret)
+        ret = strcmp(a->provider_version, b->provider_version);
+    return ret;
+}
+
+static int cmp_provider_names(const void *a_, const void *b_)
+{
+    const ap_list_provider_names_t *a = a_, *b = b_;
+    return strcmp(a->provider_name, b->provider_name);
+}
+
+static void show_providers(request_rec *r)
+{
+    apr_array_header_t *groups = ap_list_provider_groups(r->pool);
+    ap_list_provider_groups_t *group;
+    apr_array_header_t *names;
+    ap_list_provider_names_t *name;
+    int i,j;
+    const char *cur_group = NULL;
+
+    qsort(groups->elts, groups->nelts, sizeof(ap_list_provider_groups_t),
+          cmp_provider_groups);
+    ap_rputs("<h2><a name=\"providers\">Providers</a></h2>\n<dl>", r);
+
+    for (i = 0; i < groups->nelts; i++) {
+        group = &APR_ARRAY_IDX(groups, i, ap_list_provider_groups_t);
+        if (!cur_group || strcmp(cur_group, group->provider_group) != 0) {
+            if (cur_group)
+                ap_rputs("\n</dt>\n", r);
+            cur_group = group->provider_group;
+            ap_rprintf(r, "<dt><strong>%s</strong> (version <tt>%s</tt>):"
+                          "\n <br />\n", cur_group, group->provider_version);
+        }
+        names = ap_list_provider_names(r->pool, group->provider_group,
+                                       group->provider_version);
+        qsort(names->elts, names->nelts, sizeof(ap_list_provider_names_t),
+              cmp_provider_names);
+        for (j = 0; j < names->nelts; j++) {
+            name = &APR_ARRAY_IDX(names, j, ap_list_provider_names_t);
+            ap_rprintf(r, "<tt>&nbsp;&nbsp;%s</tt><br/>", name->provider_name);
+        }
+    }
+    if (cur_group)
+        ap_rputs("\n</dt>\n", r);
+    ap_rputs("</dl>\n<hr />\n", r);
+}
+
+static int cmp_module_name(const void *a_, const void *b_)
+{
+    const module * const *a = a_;
+    const module * const *b = b_;
+    return strcmp((*a)->name, (*b)->name);
+}
+
+static apr_array_header_t *get_sorted_modules(apr_pool_t *p)
+{
+    apr_array_header_t *arr = apr_array_make(p, 64, sizeof(module *));
+    module *modp, **entry;
+    for (modp = ap_top_module; modp; modp = modp->next) {
+        entry = &APR_ARRAY_PUSH(arr, module *);
+        *entry = modp;
+    }
+    qsort(arr->elts, arr->nelts, sizeof(module *), cmp_module_name);
+    return arr;
+}
+
 static int display_info(request_rec * r)
 {
     module *modp = NULL;
-    server_rec *serv = r->server;
     const char *more_info;
-    const command_rec *cmd = NULL;
-    int comma = 0;
+    const command_rec *cmd;
+    apr_array_header_t *modules = NULL;
+    int i;
 
-    if (strcmp(r->handler, "server-info"))
+    if (strcmp(r->handler, "server-info")) {
         return DECLINED;
+    }
 
     r->allowed |= (AP_METHOD_BIT << M_GET);
-    if (r->method_number != M_GET)
+    if (r->method_number != M_GET) {
         return DECLINED;
+    }
 
     ap_set_content_type(r, "text/html; charset=ISO-8859-1");
 
@@ -620,22 +790,29 @@ static int display_info(request_rec * r)
             ap_rputs("<dl><dt><tt>Subpages:<br />", r);
             ap_rputs("<a href=\"?config\">Configuration Files</a>, "
                      "<a href=\"?server\">Server Settings</a>, "
-                     "<a href=\"?list\">Module List</a>,  "
-                     "<a href=\"?hooks\">Active Hooks</a>", r);
+                     "<a href=\"?list\">Module List</a>, "
+                     "<a href=\"?hooks\">Active Hooks</a>, "
+                     "<a href=\"?providers\">Available Providers</a>", r);
             ap_rputs("</tt></dt></dl><hr />", r);
 
             ap_rputs("<dl><dt><tt>Sections:<br />", r);
-            ap_rputs("<a href=\"#server\">Server Settings</a>, "
+            ap_rputs("<a href=\"#modules\">Loaded Modules</a>, "
+                     "<a href=\"#server\">Server Settings</a>, "
                      "<a href=\"#startup_hooks\">Startup Hooks</a>, "
-                     "<a href=\"#request_hooks\">Request Hooks</a>", r);
+                     "<a href=\"#request_hooks\">Request Hooks</a>, "
+                     "<a href=\"#other_hooks\">Other Hooks</a>, "
+                     "<a href=\"#providers\">Providers</a>", r);
             ap_rputs("</tt></dt></dl><hr />", r);
 
-            ap_rputs("<dl><dt><tt>Loaded Modules: <br />", r);
-            /* TODO: Sort by Alpha */
-            for (modp = ap_top_module; modp; modp = modp->next) {
+            ap_rputs("<h2><a name=\"modules\">Loaded Modules</a></h2>"
+                    "<dl><dt><tt>", r);
+
+            modules = get_sorted_modules(r->pool);
+            for (i = 0; i < modules->nelts; i++) {
+                modp = APR_ARRAY_IDX(modules, i, module *);
                 ap_rprintf(r, "<a href=\"#%s\">%s</a>", modp->name,
                            modp->name);
-                if (modp->next) {
+                if (i < modules->nelts) {
                     ap_rputs(", ", r);
                 }
             }
@@ -650,13 +827,21 @@ static int display_info(request_rec * r)
             show_active_hooks(r);
         }
 
+        if (!r->args || !strcasecmp(r->args, "providers")) {
+            show_providers(r);
+        }
+
         if (r->args && 0 == strcasecmp(r->args, "config")) {
             ap_rputs("<dl><dt><strong>Configuration:</strong>\n", r);
             mod_info_module_cmds(r, NULL, ap_conftree, 0, 0);
             ap_rputs("</dl><hr />", r);
         }
         else {
-            for (modp = ap_top_module; modp; modp = modp->next) {
+            int comma = 0;
+            if (!modules)
+                 modules = get_sorted_modules(r->pool);
+            for (i = 0; i < modules->nelts; i++) {
+                modp = APR_ARRAY_IDX(modules, i, module *);
                 if (!r->args || !strcasecmp(modp->name, r->args)) {
                     ap_rprintf(r,
                                "<dl><dt><a name=\"%s\"><strong>Module Name:</strong></a> "
@@ -741,7 +926,7 @@ static int display_info(request_rec * r)
                             ("<dt><strong>Module Directives:</strong> <tt>none</tt></dt>",
                              r);
                     }
-                    more_info = find_more_info(serv, modp->name);
+                    more_info = find_more_info(r->server, modp->name);
                     if (more_info) {
                         ap_rputs
                             ("<dt><strong>Additional Information:</strong>\n</dt><dd>",
@@ -762,7 +947,9 @@ static int display_info(request_rec * r)
     }
     else {
         ap_rputs("<dl><dt>Server Module List</dt>", r);
-        for (modp = ap_top_module; modp; modp = modp->next) {
+        modules = get_sorted_modules(r->pool);
+        for (i = 0; i < modules->nelts; i++) {
+            modp = APR_ARRAY_IDX(modules, i, module *);
             ap_rputs("<dd>", r);
             ap_rputs(modp->name, r);
             ap_rputs("</dd>", r);
@@ -795,12 +982,25 @@ static const command_rec info_cmds[] = {
     {NULL}
 };
 
+static int check_config(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp,
+                        server_rec *s)
+{
+    if (ap_exists_config_define("DUMP_CONFIG")) {
+        apr_file_open_stdout(&out, ptemp);
+        mod_info_module_cmds(NULL, NULL, ap_conftree, 0, 0);
+    }
+
+    return DECLINED;
+}
+
+
 static void register_hooks(apr_pool_t * p)
 {
     ap_hook_handler(display_info, NULL, NULL, APR_HOOK_MIDDLE);
+    ap_hook_check_config(check_config, NULL, NULL, APR_HOOK_FIRST);
 }
 
-module AP_MODULE_DECLARE_DATA info_module = {
+AP_DECLARE_MODULE(info) = {
     STANDARD20_MODULE_STUFF,
     NULL,                       /* dir config creater */
     NULL,                       /* dir merger --- default is to override */
