@@ -34,6 +34,7 @@
 #include "http_config.h"
 #include "http_request.h"
 #include "http_log.h"
+#include "ap_expr.h"
 
 
 typedef struct {
@@ -50,11 +51,20 @@ typedef struct {
 } alias_server_conf;
 
 typedef struct {
+    unsigned int alias_set:1;
+    unsigned int redirect_set:1;
     apr_array_header_t *redirects;
+    const ap_expr_info_t *alias;
+    char *handler;
+    const ap_expr_info_t *redirect;
+    int redirect_status;                /* 301, 302, 303, 410, etc */
 } alias_dir_conf;
 
 module AP_MODULE_DECLARE_DATA alias_module;
 
+static char magic_error_value;
+#define PREGSUB_ERROR      (&magic_error_value)
+
 static void *create_alias_config(apr_pool_t *p, server_rec *s)
 {
     alias_server_conf *a =
@@ -91,7 +101,17 @@ static void *merge_alias_dir_config(apr_pool_t *p, void *basev, void *overridesv
     (alias_dir_conf *) apr_pcalloc(p, sizeof(alias_dir_conf));
     alias_dir_conf *base = (alias_dir_conf *) basev;
     alias_dir_conf *overrides = (alias_dir_conf *) overridesv;
+
     a->redirects = apr_array_append(p, overrides->redirects, base->redirects);
+
+    a->alias = (overrides->alias_set == 0) ? base->alias : overrides->alias;
+    a->handler = (overrides->alias_set == 0) ? base->handler : overrides->handler;
+    a->alias_set = overrides->alias_set || base->alias_set;
+
+    a->redirect = (overrides->redirect_set == 0) ? base->redirect : overrides->redirect;
+    a->redirect_status = (overrides->redirect_set == 0) ? base->redirect_status : overrides->redirect_status;
+    a->redirect_set = overrides->redirect_set || base->redirect_set;
+
     return a;
 }
 
@@ -99,7 +119,7 @@ static void *merge_alias_dir_config(apr_pool_t *p, void *basev, void *overridesv
 static int alias_matches(const char *uri, const char *alias_fakename);
 
 static const char *add_alias_internal(cmd_parms *cmd, void *dummy,
-                                      const char *f, const char *r,
+                                      const char *fake, const char *real,
                                       int use_regex)
 {
     server_rec *s = cmd->server;
@@ -109,13 +129,19 @@ static const char *add_alias_internal(cmd_parms *cmd, void *dummy,
     alias_entry *entries = (alias_entry *)conf->aliases->elts;
     int i;
 
-    /* XX r can NOT be relative to DocumentRoot here... compat bug. */
+    /* XXX: real can NOT be relative to DocumentRoot here... compat bug. */
+
+    const char *err = ap_check_cmd_context(cmd, NOT_IN_DIR_CONTEXT);
+
+    if (err != NULL) {
+        return err;
+    }
 
     if (use_regex) {
-        new->regexp = ap_pregcomp(cmd->pool, f, AP_REG_EXTENDED);
+        new->regexp = ap_pregcomp(cmd->pool, fake, AP_REG_EXTENDED);
         if (new->regexp == NULL)
             return "Regular expression could not be compiled.";
-        new->real = r;
+        new->real = real;
     }
     else {
         /* XXX This may be optimized, but we must know that new->real
@@ -123,9 +149,9 @@ static const char *add_alias_internal(cmd_parms *cmd, void *dummy,
          * and just canonicalizing the remainder.  Not till I finish
          * cleaning out the old ap_canonical stuff first.
          */
-        new->real = r;
+        new->real = real;
     }
-    new->fake = f;
+    new->fake = fake;
     new->handler = cmd->info;
 
     /* check for overlapping (Script)Alias directives
@@ -133,18 +159,18 @@ static const char *add_alias_internal(cmd_parms *cmd, void *dummy,
      */
     if (!use_regex) {
         for (i = 0; i < conf->aliases->nelts - 1; ++i) {
-            alias_entry *p = &entries[i];
+            alias_entry *alias = &entries[i];
 
-            if (  (!p->regexp &&  alias_matches(f, p->fake) > 0)
-                || (p->regexp && !ap_regexec(p->regexp, f, 0, NULL, 0))) {
-                ap_log_error(APLOG_MARK, APLOG_WARNING, 0, cmd->server,
+            if (  (!alias->regexp &&  alias_matches(fake, alias->fake) > 0)
+                || (alias->regexp && !ap_regexec(alias->regexp, fake, 0, NULL, 0))) {
+                ap_log_error(APLOG_MARK, APLOG_WARNING, 0, cmd->server, APLOGNO(00671)
                              "The %s directive in %s at line %d will probably "
                              "never match because it overlaps an earlier "
                              "%sAlias%s.",
                              cmd->cmd->name, cmd->directive->filename,
                              cmd->directive->line_num,
-                             p->handler ? "Script" : "",
-                             p->regexp ? "Match" : "");
+                             alias->handler ? "Script" : "",
+                             alias->regexp ? "Match" : "");
 
                 break; /* one warning per alias should be sufficient */
             }
@@ -154,16 +180,48 @@ static const char *add_alias_internal(cmd_parms *cmd, void *dummy,
     return NULL;
 }
 
-static const char *add_alias(cmd_parms *cmd, void *dummy, const char *f,
-                             const char *r)
+static const char *add_alias(cmd_parms *cmd, void *dummy, const char *fake,
+        const char *real)
 {
-    return add_alias_internal(cmd, dummy, f, r, 0);
+    if (real) {
+
+        return add_alias_internal(cmd, dummy, fake, real, 0);
+
+    }
+    else {
+        alias_dir_conf *dirconf = (alias_dir_conf *) dummy;
+
+        const char *err = ap_check_cmd_context(cmd, NOT_IN_DIRECTORY|NOT_IN_FILES);
+
+        if (err != NULL) {
+            return err;
+        }
+
+        if (!cmd->path) {
+            return "Alias must have two arguments when used globally";
+        }
+
+        dirconf->alias =
+                ap_expr_parse_cmd(cmd, fake, AP_EXPR_FLAG_STRING_RESULT,
+                        &err, NULL);
+        if (err) {
+            return apr_pstrcat(cmd->temp_pool,
+                    "Cannot parse alias expression '", fake, "': ", err,
+                    NULL);
+        }
+
+        dirconf->handler = cmd->info;
+        dirconf->alias_set = 1;
+
+        return NULL;
+
+    }
 }
 
-static const char *add_alias_regex(cmd_parms *cmd, void *dummy, const char *f,
-                                   const char *r)
+static const char *add_alias_regex(cmd_parms *cmd, void *dummy,
+                                   const char *fake, const char *real)
 {
-    return add_alias_internal(cmd, dummy, f, r, 1);
+    return add_alias_internal(cmd, dummy, fake, real, 1);
 }
 
 static const char *add_redirect_internal(cmd_parms *cmd,
@@ -177,8 +235,8 @@ static const char *add_redirect_internal(cmd_parms *cmd,
                                                          &alias_module);
     int status = (int) (long) cmd->info;
     int grokarg1 = 1;
-    ap_regex_t *r = NULL;
-    const char *f = arg2;
+    ap_regex_t *regex = NULL;
+    const char *fake = arg2;
     const char *url = arg3;
 
     /*
@@ -194,16 +252,64 @@ static const char *add_redirect_internal(cmd_parms *cmd,
         status = HTTP_MOVED_TEMPORARILY;
     else if (!strcasecmp(arg1, "seeother"))
         status = HTTP_SEE_OTHER;
-    else if (!strcasecmp(arg1, "gone"))
+    else if (!strcasecmp(arg1, "gone")) {
         status = HTTP_GONE;
-    else if (apr_isdigit(*arg1))
+        grokarg1 = -1;
+    }
+    else if (apr_isdigit(*arg1)) {
         status = atoi(arg1);
-    else
+        if (!ap_is_HTTP_REDIRECT(status)) {
+            grokarg1 = -1;
+        }
+    }
+    else {
         grokarg1 = 0;
+    }
 
     if (arg3 && !grokarg1)
         return "Redirect: invalid first argument (of three)";
 
+    /*
+     * if we have the 2nd arg and we understand the 1st one as a redirect
+     * status (3xx, but not things like 404 /robots.txt), or if we have the
+     * 1st arg but don't understand it, we use the expression syntax assuming
+     * a path from the location.
+     *
+     * if we understand the first arg but have no second arg, we are dealing
+     * with a status like "GONE" or a non-redirect status (e.g. 404, 503).
+     */
+    if (!cmd->path) {
+        /* <Location> context only for now */
+        ;
+    }
+    else if ((grokarg1 > 0 && arg2 && !arg3) || (!grokarg1 && !arg2)) {
+        const char *expr_err = NULL;
+
+        url = grokarg1 ? arg2 : arg1;
+        dirconf->redirect =
+                ap_expr_parse_cmd(cmd, url, AP_EXPR_FLAG_STRING_RESULT,
+                        &expr_err, NULL);
+        if (expr_err) {
+            return apr_pstrcat(cmd->temp_pool,
+                    "Cannot parse redirect expression '", url, "': ", expr_err,
+                    NULL);
+        }
+
+        dirconf->redirect_status = status;
+        dirconf->redirect_set = 1;
+
+        return NULL;
+
+    }
+    else if (grokarg1 < 0 && !arg2) {
+
+        dirconf->redirect_status = status;
+        dirconf->redirect_set = 1;
+
+        return NULL;
+
+    }
+
     /*
      * if we don't have the 3rd arg and we didn't understand the 1st
      * one, then assume URL-path URL. This also handles case, eg, GONE
@@ -211,13 +317,13 @@ static const char *add_redirect_internal(cmd_parms *cmd,
      * one, so we don't want to re-arrange
      */
     if (!arg3 && !grokarg1) {
-        f = arg1;
+        fake = arg1;
         url = arg2;
     }
 
     if (use_regex) {
-        r = ap_pregcomp(cmd->pool, f, AP_REG_EXTENDED);
-        if (r == NULL)
+        regex = ap_pregcomp(cmd->pool, fake, AP_REG_EXTENDED);
+        if (regex == NULL)
             return "Regular expression could not be compiled.";
     }
 
@@ -240,9 +346,9 @@ static const char *add_redirect_internal(cmd_parms *cmd,
     else
         new = apr_array_push(serverconf->redirects);
 
-    new->fake = f;
+    new->fake = fake;
     new->real = url;
-    new->regexp = r;
+    new->regexp = regex;
     new->redir_status = status;
     return NULL;
 }
@@ -269,11 +375,11 @@ static const char *add_redirect_regex(cmd_parms *cmd, void *dirconf,
 
 static const command_rec alias_cmds[] =
 {
-    AP_INIT_TAKE2("Alias", add_alias, NULL, RSRC_CONF,
-                  "a fakename and a realname"),
-    AP_INIT_TAKE2("ScriptAlias", add_alias, "cgi-script", RSRC_CONF,
-                  "a fakename and a realname"),
-    AP_INIT_TAKE23("Redirect", add_redirect, (void *) HTTP_MOVED_TEMPORARILY,
+    AP_INIT_TAKE12("Alias", add_alias, NULL, RSRC_CONF | ACCESS_CONF,
+                  "a fakename and a realname, or a realname in a Location"),
+    AP_INIT_TAKE12("ScriptAlias", add_alias, "cgi-script", RSRC_CONF | ACCESS_CONF,
+                  "a fakename and a realname, or a realname in a Location"),
+    AP_INIT_TAKE123("Redirect", add_redirect, (void *) HTTP_MOVED_TEMPORARILY,
                    OR_FILEINFO,
                    "an optional status, then document to be redirected and "
                    "destination URL"),
@@ -333,8 +439,82 @@ static int alias_matches(const char *uri, const char *alias_fakename)
     return urip - uri;
 }
 
+static char *try_alias(request_rec *r)
+{
+    alias_dir_conf *dirconf =
+            (alias_dir_conf *) ap_get_module_config(r->per_dir_config, &alias_module);
+
+    if (dirconf->alias) {
+        const char *err = NULL;
+
+        char *found = apr_pstrdup(r->pool,
+                ap_expr_str_exec(r, dirconf->alias, &err));
+        if (err) {
+            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(02825)
+                          "Can't evaluate alias expression: %s", err);
+            return PREGSUB_ERROR;
+        }
+
+        if (dirconf->handler) { /* Set handler, and leave a note for mod_cgi */
+            r->handler = dirconf->handler;
+            apr_table_setn(r->notes, "alias-forced-type", r->handler);
+        }
+        /* XXX This is as SLOW as can be, next step, we optimize
+         * and merge to whatever part of the found path was already
+         * canonicalized.  After I finish eliminating os canonical.
+         * Better fail test for ap_server_root_relative needed here.
+         */
+        found = ap_server_root_relative(r->pool, found);
+        return found;
+
+    }
+
+    return NULL;
+}
+
+static char *try_redirect(request_rec *r, int *status)
+{
+    alias_dir_conf *dirconf =
+            (alias_dir_conf *) ap_get_module_config(r->per_dir_config, &alias_module);
+
+    if (dirconf->redirect_set) {
+        apr_uri_t uri;
+        const char *err = NULL;
+        char *found = "";
+
+        if (dirconf->redirect) {
+
+            found = apr_pstrdup(r->pool,
+                    ap_expr_str_exec(r, dirconf->redirect, &err));
+            if (err) {
+                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(02826)
+                              "Can't evaluate redirect expression: %s", err);
+                return PREGSUB_ERROR;
+            }
+
+            apr_uri_parse(r->pool, found, &uri);
+            /* Do not escape the query string or fragment. */
+            found = apr_uri_unparse(r->pool, &uri, APR_URI_UNP_OMITQUERY);
+            found = ap_escape_uri(r->pool, found);
+            if (uri.query) {
+                found = apr_pstrcat(r->pool, found, "?", uri.query, NULL);
+            }
+            if (uri.fragment) {
+                found = apr_pstrcat(r->pool, found, "#", uri.fragment, NULL);
+            }
+
+        }
+
+        *status = dirconf->redirect_status;
+        return found;
+
+    }
+
+    return NULL;
+}
+
 static char *try_alias_list(request_rec *r, apr_array_header_t *aliases,
-                            int doesc, int *status)
+                            int is_redir, int *status)
 {
     alias_entry *entries = (alias_entry *) aliases->elts;
     ap_regmatch_t regm[AP_MAX_REG_MATCH];
@@ -342,55 +522,64 @@ static char *try_alias_list(request_rec *r, apr_array_header_t *aliases,
     int i;
 
     for (i = 0; i < aliases->nelts; ++i) {
-        alias_entry *p = &entries[i];
+        alias_entry *alias = &entries[i];
         int l;
 
-        if (p->regexp) {
-            if (!ap_regexec(p->regexp, r->uri, AP_MAX_REG_MATCH, regm, 0)) {
-                if (p->real) {
-                    found = ap_pregsub(r->pool, p->real, r->uri,
+        if (alias->regexp) {
+            if (!ap_regexec(alias->regexp, r->uri, AP_MAX_REG_MATCH, regm, 0)) {
+                if (alias->real) {
+                    found = ap_pregsub(r->pool, alias->real, r->uri,
                                        AP_MAX_REG_MATCH, regm);
-                    if (found && doesc) {
-                        apr_uri_t uri;
-                        apr_uri_parse(r->pool, found, &uri);
-                        /* Do not escape the query string or fragment. */
-                        found = apr_uri_unparse(r->pool, &uri,
-                                                APR_URI_UNP_OMITQUERY);
-                        found = ap_escape_uri(r->pool, found);
-                        if (uri.query) {
-                            found = apr_pstrcat(r->pool, found, "?",
-                                                uri.query, NULL);
-                        }
-                        if (uri.fragment) {
-                            found = apr_pstrcat(r->pool, found, "#",
-                                                uri.fragment, NULL);
-                        }
+                    if (found) {
+                       if (is_redir) {
+                            apr_uri_t uri;
+                            apr_uri_parse(r->pool, found, &uri);
+                            /* Do not escape the query string or fragment. */
+                            found = apr_uri_unparse(r->pool, &uri,
+                                                    APR_URI_UNP_OMITQUERY);
+                            found = ap_escape_uri(r->pool, found);
+                            if (uri.query) {
+                                found = apr_pstrcat(r->pool, found, "?",
+                                                    uri.query, NULL);
+                            }
+                            if (uri.fragment) {
+                                found = apr_pstrcat(r->pool, found, "#",
+                                                    uri.fragment, NULL);
+                            }
+                       }
+                    }
+                    else {
+                        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(00672)
+                                      "Regex substitution in '%s' failed. "
+                                      "Replacement too long?", alias->real);
+                        return PREGSUB_ERROR;
                     }
                 }
                 else {
                     /* need something non-null */
-                    found = apr_pstrdup(r->pool, "");
+                    found = "";
                 }
             }
         }
         else {
-            l = alias_matches(r->uri, p->fake);
+            l = alias_matches(r->uri, alias->fake);
 
             if (l > 0) {
-                if (doesc) {
+                ap_set_context_info(r, alias->fake, alias->real);
+                if (is_redir) {
                     char *escurl;
                     escurl = ap_os_escape_path(r->pool, r->uri + l, 1);
 
-                    found = apr_pstrcat(r->pool, p->real, escurl, NULL);
+                    found = apr_pstrcat(r->pool, alias->real, escurl, NULL);
                 }
                 else
-                    found = apr_pstrcat(r->pool, p->real, r->uri + l, NULL);
+                    found = apr_pstrcat(r->pool, alias->real, r->uri + l, NULL);
             }
         }
 
         if (found) {
-            if (p->handler) {    /* Set handler, and leave a note for mod_cgi */
-                r->handler = p->handler;
+            if (alias->handler) {    /* Set handler, and leave a note for mod_cgi */
+                r->handler = alias->handler;
                 apr_table_setn(r->notes, "alias-forced-type", r->handler);
             }
             /* XXX This is as SLOW as can be, next step, we optimize
@@ -398,11 +587,11 @@ static char *try_alias_list(request_rec *r, apr_array_header_t *aliases,
              * canonicalized.  After I finish eliminating os canonical.
              * Better fail test for ap_server_root_relative needed here.
              */
-            if (!doesc) {
+            if (!is_redir) {
                 found = ap_server_root_relative(r->pool, found);
             }
             if (found) {
-                *status = p->redir_status;
+                *status = alias->redir_status;
             }
             return found;
         }
@@ -423,39 +612,44 @@ static int translate_alias_redir(request_rec *r)
         return DECLINED;
     }
 
-    if ((ret = try_alias_list(r, serverconf->redirects, 1, &status)) != NULL) {
+    if ((ret = try_redirect(r, &status)) != NULL
+            || (ret = try_alias_list(r, serverconf->redirects, 1, &status))
+                    != NULL) {
+        if (ret == PREGSUB_ERROR)
+            return HTTP_INTERNAL_SERVER_ERROR;
         if (ap_is_HTTP_REDIRECT(status)) {
-            char *orig_target = ret;
             if (ret[0] == '/') {
+                char *orig_target = ret;
 
                 ret = ap_construct_url(r->pool, ret, r);
-                ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
+                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(00673)
                               "incomplete redirection target of '%s' for "
                               "URI '%s' modified to '%s'",
                               orig_target, r->uri, ret);
             }
             if (!ap_is_url(ret)) {
-                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+                status = HTTP_INTERNAL_SERVER_ERROR;
+                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(00674)
                               "cannot redirect '%s' to '%s'; "
                               "target is not a valid absoluteURI or abs_path",
                               r->uri, ret);
-                /* restore the config value, so as not to get a
-                 * "regression" on existing "working" configs.
+            }
+            else {
+                /* append requested query only, if the config didn't
+                 * supply its own.
                  */
-                ret = orig_target;
+                if (r->args && !ap_strchr(ret, '?')) {
+                    ret = apr_pstrcat(r->pool, ret, "?", r->args, NULL);
+                }
+                apr_table_setn(r->headers_out, "Location", ret);
             }
-            /* append requested query only, if the config didn't
-             * supply its own.
-             */
-            if (r->args && !ap_strchr(ret, '?')) {
-                ret = apr_pstrcat(r->pool, ret, "?", r->args, NULL);
-            }
-            apr_table_setn(r->headers_out, "Location", ret);
         }
         return status;
     }
 
-    if ((ret = try_alias_list(r, serverconf->aliases, 0, &status)) != NULL) {
+    if ((ret = try_alias(r)) != NULL
+            || (ret = try_alias_list(r, serverconf->aliases, 0, &status))
+                    != NULL) {
         r->filename = ret;
         return OK;
     }
@@ -473,20 +667,24 @@ static int fixup_redir(request_rec *r)
 
     /* It may have changed since last time, so try again */
 
-    if ((ret = try_alias_list(r, dirconf->redirects, 1, &status)) != NULL) {
+    if ((ret = try_redirect(r, &status)) != NULL
+            || (ret = try_alias_list(r, dirconf->redirects, 1, &status))
+                    != NULL) {
+        if (ret == PREGSUB_ERROR)
+            return HTTP_INTERNAL_SERVER_ERROR;
         if (ap_is_HTTP_REDIRECT(status)) {
             if (ret[0] == '/') {
                 char *orig_target = ret;
 
                 ret = ap_construct_url(r->pool, ret, r);
-                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(00675)
                               "incomplete redirection target of '%s' for "
                               "URI '%s' modified to '%s'",
                               orig_target, r->uri, ret);
             }
             if (!ap_is_url(ret)) {
                 status = HTTP_INTERNAL_SERVER_ERROR;
-                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(00676)
                               "cannot redirect '%s' to '%s'; "
                               "target is not a valid absoluteURI or abs_path",
                               r->uri, ret);
@@ -516,7 +714,7 @@ static void register_hooks(apr_pool_t *p)
     ap_hook_fixups(fixup_redir,NULL,NULL,APR_HOOK_MIDDLE);
 }
 
-module AP_MODULE_DECLARE_DATA alias_module =
+AP_DECLARE_MODULE(alias) =
 {
     STANDARD20_MODULE_STUFF,
     create_alias_dir_config,       /* dir config creater */
