@@ -81,6 +81,7 @@
 #include "http_log.h"
 #include "util_filter.h"
 #include "http_protocol.h"
+#include "ap_expr.h"
 
 #include "mod_ssl.h" /* for the ssl_var_lookup optional function defn */
 
@@ -94,22 +95,28 @@ typedef enum {
     hdr_merge = 'g',            /* merge (merge, but avoid duplicates) */
     hdr_unset = 'u',            /* unset header */
     hdr_echo = 'e',             /* echo headers from request to response */
-    hdr_edit = 'r'              /* change value by regexp */
+    hdr_edit = 'r',             /* change value by regexp, match once */
+    hdr_edit_r = 'R',           /* change value by regexp, everymatch */
+    hdr_setifempty = 'i',       /* set value if header not already present*/
+    hdr_note = 'n'              /* set value of header in a note */
 } hdr_actions;
 
 /*
  * magic cmd->info values
  */
 static char hdr_in  = '0';  /* RequestHeader */
-static char hdr_out = '1';  /* Header onsuccess */
-static char hdr_err = '2';  /* Header always */
+static char hdr_out_onsuccess = '1';  /* Header onsuccess */
+static char hdr_out_always = '2';  /* Header always */
+
+/* Callback function type. */
+typedef const char *format_tag_fn(request_rec *r, char *a);
 
 /*
  * There is an array of struct format_tag per Header/RequestHeader
  * config directive
  */
 typedef struct {
-    const char* (*func)(request_rec *r,char *arg);
+    format_tag_fn *func;
     char *arg;
 } format_tag;
 
@@ -125,6 +132,8 @@ typedef struct {
     ap_regex_t *regex;
     const char *condition_var;
     const char *subs;
+    ap_expr_info_t *expr;
+    ap_expr_info_t *expr_out;
 } header_entry;
 
 /* echo_do is used for Header echo to iterate through the request headers*/
@@ -135,7 +144,7 @@ typedef struct {
 
 /* edit_do is used for Header edit to iterate through the request headers */
 typedef struct {
-    apr_pool_t *p;
+    request_rec *r;
     header_entry *hdr;
     apr_table_t *t;
 } edit_do;
@@ -214,6 +223,28 @@ static const char *header_request_ssl_var(request_rec *r, char *name)
     }
 }
 
+static const char *header_request_loadavg(request_rec *r, char *a)
+{
+    ap_loadavg_t t;
+    ap_get_loadavg(&t);
+    return apr_psprintf(r->pool, "l=%.2f/%.2f/%.2f", t.loadavg,
+                        t.loadavg5, t.loadavg15);
+}
+
+static const char *header_request_idle(request_rec *r, char *a)
+{
+    ap_sload_t t;
+    ap_get_sload(&t);
+    return apr_psprintf(r->pool, "i=%d", t.idle);
+}
+
+static const char *header_request_busy(request_rec *r, char *a)
+{
+    ap_sload_t t;
+    ap_get_sload(&t);
+    return apr_psprintf(r->pool, "b=%d", t.busy);
+}
+
 /*
  * Config routines
  */
@@ -326,7 +357,7 @@ static char *parse_format_tag(apr_pool_t *p, format_tag *tag, const char **sa)
         return NULL;
     }
 
-    tag->arg = '\0';
+    tag->arg = "\0";
     /* grab the argument if there is one */
     if (*s == '{') {
         ++s;
@@ -357,14 +388,29 @@ static char *parse_format_tag(apr_pool_t *p, format_tag *tag, const char **sa)
  * contains a pointer to the function used to format the tag. Then save each
  * tag in the tag array anchored in the header_entry.
  */
-static char *parse_format_string(apr_pool_t *p, header_entry *hdr, const char *s)
+static char *parse_format_string(cmd_parms *cmd, header_entry *hdr, const char *s)
 {
+    apr_pool_t *p = cmd->pool;
     char *res;
 
     /* No string to parse with unset and echo commands */
-    if (hdr->action == hdr_unset ||
-        hdr->action == hdr_edit ||
-        hdr->action == hdr_echo) {
+    if (hdr->action == hdr_unset || hdr->action == hdr_echo) {
+        return NULL;
+    }
+    /* Tags are in the replacement value for edit */
+    else if (hdr->action == hdr_edit || hdr->action == hdr_edit_r ) {
+        s = hdr->subs;
+    }
+
+    if (!strncmp(s, "expr=", 5)) { 
+        const char *err;
+        hdr->expr_out = ap_expr_parse_cmd(cmd, s+5, 
+                                          AP_EXPR_FLAG_STRING_RESULT,
+                                          &err, NULL);
+        if (err) {
+            return apr_pstrcat(cmd->pool,
+                    "Can't parse value expression : ", err, NULL);
+        }
         return NULL;
     }
 
@@ -391,9 +437,10 @@ static APR_INLINE const char *header_inout_cmd(cmd_parms *cmd,
     const char *condition_var = NULL;
     const char *colon;
     header_entry *new;
+    ap_expr_info_t *expr = NULL;
 
     apr_array_header_t *fixup = (cmd->info == &hdr_in)
-        ? dirconf->fixup_in   : (cmd->info == &hdr_err)
+        ? dirconf->fixup_in   : (cmd->info == &hdr_out_always)
         ? dirconf->fixup_err
         : dirconf->fixup_out;
 
@@ -401,6 +448,8 @@ static APR_INLINE const char *header_inout_cmd(cmd_parms *cmd,
 
     if (!strcasecmp(action, "set"))
         new->action = hdr_set;
+    else if (!strcasecmp(action, "setifempty"))
+        new->action = hdr_setifempty;
     else if (!strcasecmp(action, "add"))
         new->action = hdr_add;
     else if (!strcasecmp(action, "append"))
@@ -413,11 +462,15 @@ static APR_INLINE const char *header_inout_cmd(cmd_parms *cmd,
         new->action = hdr_echo;
     else if (!strcasecmp(action, "edit"))
         new->action = hdr_edit;
+    else if (!strcasecmp(action, "edit*"))
+        new->action = hdr_edit_r;
+    else if (!strcasecmp(action, "note"))
+        new->action = hdr_note;
     else
-        return "first argument must be 'add', 'set', 'append', 'merge', "
-               "'unset', 'echo', or 'edit'.";
+        return "first argument must be 'add', 'set', 'setifempty', 'append', 'merge', "
+               "'unset', 'echo', 'note', 'edit', or 'edit*'.";
 
-    if (new->action == hdr_edit) {
+    if (new->action == hdr_edit || new->action == hdr_edit_r) {
         if (subs == NULL) {
             return "Header edit requires a match and a substitution";
         }
@@ -453,7 +506,7 @@ static APR_INLINE const char *header_inout_cmd(cmd_parms *cmd,
             envclause = value;
             value = NULL;
         }
-        if (cmd->info != &hdr_out && cmd->info != &hdr_err)
+        if (cmd->info != &hdr_out_onsuccess && cmd->info != &hdr_out_always)
             return "Header echo only valid on Header "
                    "directives";
         else {
@@ -472,10 +525,7 @@ static APR_INLINE const char *header_inout_cmd(cmd_parms *cmd,
         if (strcasecmp(envclause, "early") == 0) {
             condition_var = condition_early;
         }
-        else {
-            if (strncasecmp(envclause, "env=", 4) != 0) {
-                return "error: envclause should be in the form env=envar";
-            }
+        else if (strncasecmp(envclause, "env=", 4) == 0) {
             if ((envclause[4] == '\0')
                 || ((envclause[4] == '!') && (envclause[5] == '\0'))) {
                 return "error: missing environment variable name. "
@@ -483,6 +533,19 @@ static APR_INLINE const char *header_inout_cmd(cmd_parms *cmd,
             }
             condition_var = envclause + 4;
         }
+        else if (strncasecmp(envclause, "expr=", 5) == 0) {
+            const char *err = NULL;
+            expr = ap_expr_parse_cmd(cmd, envclause + 5, 0, &err, NULL);
+            if (err) {
+                return apr_pstrcat(cmd->pool,
+                                   "Can't parse envclause/expression: ", err,
+                                   NULL);
+            }
+        }
+        else {
+            return apr_pstrcat(cmd->pool, "Unknown parameter: ", envclause,
+                               NULL);
+        }
     }
 
     if ((colon = ap_strchr_c(hdr, ':'))) {
@@ -491,8 +554,9 @@ static APR_INLINE const char *header_inout_cmd(cmd_parms *cmd,
 
     new->header = hdr;
     new->condition_var = condition_var;
+    new->expr = expr;
 
-    return parse_format_string(cmd->pool, new, value);
+    return parse_format_string(cmd, new, value);
 }
 
 /* Handle all (xxx)Header directives */
@@ -505,14 +569,14 @@ static const char *header_cmd(cmd_parms *cmd, void *indirconf,
     const char *envclause;
     const char *subs;
 
-    action = ap_getword_conf(cmd->pool, &args);
-    if (cmd->info == &hdr_out) {
+    action = ap_getword_conf(cmd->temp_pool, &args);
+    if (cmd->info == &hdr_out_onsuccess) {
         if (!strcasecmp(action, "always")) {
-            cmd->info = &hdr_err;
-            action = ap_getword_conf(cmd->pool, &args);
+            cmd->info = &hdr_out_always;
+            action = ap_getword_conf(cmd->temp_pool, &args);
         }
         else if (!strcasecmp(action, "onsuccess")) {
-            action = ap_getword_conf(cmd->pool, &args);
+            action = ap_getword_conf(cmd->temp_pool, &args);
         }
     }
     hdr = ap_getword_conf(cmd->pool, &args);
@@ -534,14 +598,29 @@ static const char *header_cmd(cmd_parms *cmd, void *indirconf,
  * (formatter) specific to the tag. Handlers return text strings.
  * Concatenate the return from each handler into one string that is
  * returned from this call.
+ * If the original value was prefixed with "expr=", processing is
+ * handled instead by ap_expr.
  */
 static char* process_tags(header_entry *hdr, request_rec *r)
 {
     int i;
     const char *s;
     char *str = NULL;
+    format_tag *tag = NULL;
 
-    format_tag *tag = (format_tag*) hdr->ta->elts;
+    if (hdr->expr_out) { 
+        const char *err;
+        const char *val;
+        val = ap_expr_str_exec(r, hdr->expr_out, &err);
+        if (err) { 
+            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(02557)
+                          "Can't evaluate value expression: %s", err);
+            return "";
+        }
+        return apr_pstrdup(r->pool, val);
+    }
+
+    tag = (format_tag*) hdr->ta->elts;
 
     for (i = 0; i < hdr->ta->nelts; i++) {
         s = tag[i].func(r, tag[i].arg);
@@ -553,33 +632,49 @@ static char* process_tags(header_entry *hdr, request_rec *r)
     return str ? str : "";
 }
 static const char *process_regexp(header_entry *hdr, const char *value,
-                                  apr_pool_t *pool)
+                                  request_rec *r)
 {
-    unsigned int nmatch = 10;
-    ap_regmatch_t pmatch[10];
+    ap_regmatch_t pmatch[AP_MAX_REG_MATCH];
     const char *subs;
+    const char *remainder;
     char *ret;
     int diffsz;
-    if (ap_regexec(hdr->regex, value, nmatch, pmatch, 0)) {
+    if (ap_regexec(hdr->regex, value, AP_MAX_REG_MATCH, pmatch, 0)) {
         /* no match, nothing to do */
         return value;
     }
-    subs = ap_pregsub(pool, hdr->subs, value, nmatch, pmatch);
+    /* Process tags in the input string rather than the resulting
+       * substitution to avoid surprises
+       */
+    subs = ap_pregsub(r->pool, process_tags(hdr, r), value, AP_MAX_REG_MATCH, pmatch);
+    if (subs == NULL)
+        return NULL;
     diffsz = strlen(subs) - (pmatch[0].rm_eo - pmatch[0].rm_so);
-    ret = apr_palloc(pool, strlen(value) + 1 + diffsz);
+    if (hdr->action == hdr_edit) {
+        remainder = value + pmatch[0].rm_eo;
+    }
+    else { /* recurse to edit multiple matches if applicable */
+        remainder = process_regexp(hdr, value + pmatch[0].rm_eo, r);
+        if (remainder == NULL)
+            return NULL;
+        diffsz += strlen(remainder) - strlen(value + pmatch[0].rm_eo);
+    }
+    ret = apr_palloc(r->pool, strlen(value) + 1 + diffsz);
     memcpy(ret, value, pmatch[0].rm_so);
     strcpy(ret + pmatch[0].rm_so, subs);
-    strcat(ret, value + pmatch[0].rm_eo);
+    strcat(ret, remainder);
     return ret;
 }
 
-static int echo_header(echo_do *v, const char *key, const char *val)
+static int echo_header(void *v, const char *key, const char *val)
 {
+    echo_do *ed = (echo_do *)v;
+
     /* If the input header (key) matches the regex, echo it intact to
      * r->headers_out.
      */
-    if (!ap_regexec(v->hdr->regex, key, 0, NULL, 0)) {
-        apr_table_add(v->r->headers_out, key, val);
+    if (!ap_regexec(ed->hdr->regex, key, 0, NULL, 0)) {
+        apr_table_add(ed->r->headers_out, key, val);
     }
 
     return 1;
@@ -588,8 +683,11 @@ static int echo_header(echo_do *v, const char *key, const char *val)
 static int edit_header(void *v, const char *key, const char *val)
 {
     edit_do *ed = (edit_do *)v;
+    const char *repl = process_regexp(ed->hdr, val, ed->r);
+    if (repl == NULL)
+        return 0;
 
-    apr_table_addn(ed->t, key, process_regexp(ed->hdr, val, ed->p));
+    apr_table_addn(ed->t, key, repl);
     return 1;
 }
 
@@ -601,7 +699,7 @@ static int add_them_all(void *v, const char *key, const char *val)
     return 1;
 }
 
-static void do_headers_fixup(request_rec *r, apr_table_t *headers,
+static int do_headers_fixup(request_rec *r, apr_table_t *headers,
                              apr_array_header_t *fixup, int early)
 {
     echo_do v;
@@ -620,6 +718,19 @@ static void do_headers_fixup(request_rec *r, apr_table_t *headers,
         else if (early && (envar != condition_early)) {
             continue;
         }
+        /* Do we have an expression to evaluate? */
+        else if (hdr->expr != NULL) {
+            const char *err = NULL;
+            int eval = ap_expr_exec(r, hdr->expr, &err);
+            if (err) {
+                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01501)
+                              "Failed to evaluate expression (%s) - ignoring",
+                              err);
+            }
+            else if (!eval) {
+                continue;
+            }
+        }
         /* Have any conditional envar-controlled Header processing to do? */
         else if (envar && !early) {
             if (*envar != '!') {
@@ -652,7 +763,7 @@ static void do_headers_fixup(request_rec *r, apr_table_t *headers,
                 while (*val) {
                     const char *tok_start;
 
-                    while (*val && apr_isspace(*val))
+                    while (apr_isspace(*val))
                         ++val;
 
                     tok_start = val;
@@ -685,30 +796,50 @@ static void do_headers_fixup(request_rec *r, apr_table_t *headers,
             }
             apr_table_setn(headers, hdr->header, process_tags(hdr, r));
             break;
+        case hdr_setifempty:
+            if (NULL == apr_table_get(headers, hdr->header)) {
+                if (!strcasecmp(hdr->header, "Content-Type")) {
+                    ap_set_content_type(r, process_tags(hdr, r));
+                }
+                apr_table_setn(headers, hdr->header, process_tags(hdr, r));
+            }
+            break;
         case hdr_unset:
             apr_table_unset(headers, hdr->header);
             break;
         case hdr_echo:
             v.r = r;
             v.hdr = hdr;
-            apr_table_do((int (*) (void *, const char *, const char *))
-                         echo_header, (void *) &v, r->headers_in, NULL);
+            apr_table_do(echo_header, &v, r->headers_in, NULL);
             break;
         case hdr_edit:
+        case hdr_edit_r:
+            if (!strcasecmp(hdr->header, "Content-Type") && r->content_type) {
+                const char *repl = process_regexp(hdr, r->content_type, r);
+                if (repl == NULL)
+                    return 0;
+                ap_set_content_type(r, repl);
+            }
             if (apr_table_get(headers, hdr->header)) {
                 edit_do ed;
 
-                ed.p = r->pool;
+                ed.r = r;
                 ed.hdr = hdr;
                 ed.t = apr_table_make(r->pool, 5);
-                apr_table_do(edit_header, (void *) &ed, headers, hdr->header,
-                             NULL);
+                if (!apr_table_do(edit_header, (void *) &ed, headers,
+                                  hdr->header, NULL))
+                    return 0;
                 apr_table_unset(headers, hdr->header);
                 apr_table_do(add_them_all, (void *) headers, ed.t, NULL);
             }
             break;
+        case hdr_note:
+            apr_table_setn(r->notes, process_tags(hdr, r), apr_table_get(headers, hdr->header));
+            break;
+ 
         }
     }
+    return 1;
 }
 
 static void ap_headers_insert_output_filter(request_rec *r)
@@ -740,7 +871,7 @@ static apr_status_t ap_headers_output_filter(ap_filter_t *f,
     headers_conf *dirconf = ap_get_module_config(f->r->per_dir_config,
                                                  &headers_module);
 
-    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, f->r->server,
+    ap_log_error(APLOG_MARK, APLOG_TRACE2, 0, f->r->server, APLOGNO(01502)
                  "headers: ap_headers_output_filter()");
 
     /* do the fixup */
@@ -765,7 +896,7 @@ static apr_status_t ap_headers_error_filter(ap_filter_t *f,
 
     dirconf = ap_get_module_config(f->r->per_dir_config,
                                     &headers_module);
-    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, f->r->server,
+    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, f->r->server, APLOGNO(01503)
                  "headers: ap_headers_error_filter()");
 
     /*
@@ -805,21 +936,28 @@ static apr_status_t ap_headers_early(request_rec *r)
 
     /* do the fixup */
     if (dirconf->fixup_in->nelts) {
-        do_headers_fixup(r, r->headers_in, dirconf->fixup_in, 1);
+        if (!do_headers_fixup(r, r->headers_in, dirconf->fixup_in, 1))
+            goto err;
     }
     if (dirconf->fixup_err->nelts) {
-        do_headers_fixup(r, r->err_headers_out, dirconf->fixup_err, 1);
+        if (!do_headers_fixup(r, r->err_headers_out, dirconf->fixup_err, 1))
+            goto err;
     }
     if (dirconf->fixup_out->nelts) {
-        do_headers_fixup(r, r->headers_out, dirconf->fixup_out, 1);
+        if (!do_headers_fixup(r, r->headers_out, dirconf->fixup_out, 1))
+            goto err;
     }
 
     return DECLINED;
+err:
+    ap_log_rerror(APLOG_MARK, APLOG_CRIT, 0, r, APLOGNO(01504)
+                  "Regular expression replacement failed (replacement too long?)");
+    return HTTP_INTERNAL_SERVER_ERROR;
 }
 
 static const command_rec headers_cmds[] =
 {
-    AP_INIT_RAW_ARGS("Header", header_cmd, &hdr_out, OR_FILEINFO,
+    AP_INIT_RAW_ARGS("Header", header_cmd, &hdr_out_onsuccess, OR_FILEINFO,
                      "an optional condition, an action, header and value "
                      "followed by optional env clause"),
     AP_INIT_RAW_ARGS("RequestHeader", header_cmd, &hdr_in, OR_FILEINFO,
@@ -829,7 +967,7 @@ static const command_rec headers_cmds[] =
 };
 
 static void register_format_tag_handler(const char *tag,
-                                        const void *tag_handler)
+                                        format_tag_fn *tag_handler)
 {
     apr_hash_set(format_tag_hash, tag, 1, tag_handler);
 }
@@ -837,10 +975,13 @@ static void register_format_tag_handler(const char *tag,
 static int header_pre_config(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp)
 {
     format_tag_hash = apr_hash_make(p);
-    register_format_tag_handler("D", (const void *)header_request_duration);
-    register_format_tag_handler("t", (const void *)header_request_time);
-    register_format_tag_handler("e", (const void *)header_request_env_var);
-    register_format_tag_handler("s", (const void *)header_request_ssl_var);
+    register_format_tag_handler("D", header_request_duration);
+    register_format_tag_handler("t", header_request_time);
+    register_format_tag_handler("e", header_request_env_var);
+    register_format_tag_handler("s", header_request_ssl_var);
+    register_format_tag_handler("l", header_request_loadavg);
+    register_format_tag_handler("i", header_request_idle);
+    register_format_tag_handler("b", header_request_busy);
 
     return OK;
 }
@@ -867,7 +1008,7 @@ static void register_hooks(apr_pool_t *p)
     ap_hook_post_read_request(ap_headers_early, NULL, NULL, APR_HOOK_FIRST);
 }
 
-module AP_MODULE_DECLARE_DATA headers_module =
+AP_DECLARE_MODULE(headers) =
 {
     STANDARD20_MODULE_STUFF,
     create_headers_dir_config,  /* dir config creater */
