@@ -161,7 +161,7 @@ struct magic {
         unsigned char hl[4];   /* 2 bytes of a fixed-endian "long" */
     } value;                   /* either number or string */
     unsigned long mask;        /* mask before comparison with value */
-    char nospflag;             /* supress space character */
+    char nospflag;             /* suppress space character */
 
     /* NOTE: this string is suspected of overrunning - find it! */
     char desc[MAXDESC];        /* description */
@@ -257,7 +257,7 @@ static int fsmagic(request_rec *r, const char *fn);
 #define L_MAIL    8   /* Electronic mail */
 #define L_NEWS    9   /* Usenet Netnews */
 
-static char *types[] =
+static const char *types[] =
 {
     "text/html",             /* HTML */
     "text/plain",            /* "c program text", */
@@ -273,8 +273,8 @@ static char *types[] =
     0
 };
 
-static struct names {
-    char *name;
+static const struct names {
+    const char *name;
     short type;
 } names[] = {
 
@@ -443,7 +443,7 @@ static struct names {
  */
 
 typedef struct magic_rsl_s {
-    char *str;                  /* string, possibly a fragment */
+    const char *str;                  /* string, possibly a fragment */
     struct magic_rsl_s *next;   /* pointer to next fragment */
 } magic_rsl;
 
@@ -540,7 +540,7 @@ static magic_req_rec *magic_set_config(request_rec *r)
 
 /* add a string to the result string list for this request */
 /* it is the responsibility of the caller to allocate "str" */
-static int magic_rsl_add(request_rec *r, char *str)
+static int magic_rsl_add(request_rec *r, const char *str)
 {
     magic_req_rec *req_dat = (magic_req_rec *)
                     ap_get_module_config(r->request_config, &mime_magic_module);
@@ -548,7 +548,7 @@ static int magic_rsl_add(request_rec *r, char *str)
 
     /* make sure we have a list to put it in */
     if (!req_dat) {
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, APR_EINVAL, r,
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, APR_EINVAL, r, APLOGNO(01507)
                     MODNAME ": request config should not be NULL");
         if (!(req_dat = magic_set_config(r))) {
             /* failure */
@@ -577,7 +577,7 @@ static int magic_rsl_add(request_rec *r, char *str)
 }
 
 /* RSL hook for puts-type functions */
-static int magic_rsl_puts(request_rec *r, char *str)
+static int magic_rsl_puts(request_rec *r, const char *str)
 {
     return magic_rsl_add(r, str);
 }
@@ -606,7 +606,7 @@ static int magic_rsl_putchar(request_rec *r, char c)
     /* high overhead for 1 char - just hope they don't do this much */
     str[0] = c;
     str[1] = '\0';
-    return magic_rsl_add(r, str);
+    return magic_rsl_add(r, apr_pstrdup(r->pool, str));
 }
 
 /* allocate and copy a contiguous string from a result string list */
@@ -650,7 +650,7 @@ static char *rsl_strdup(request_rec *r, int start_frag, int start_pos, int len)
     /* clean up and return */
     result[res_pos] = 0;
 #if MIME_MAGIC_DEBUG
-    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01508)
              MODNAME ": rsl_strdup() %d chars: %s", res_pos - 1, result);
 #endif
     return result;
@@ -673,6 +673,7 @@ static int magic_rsl_to_request(request_rec *r)
         encoding_pos,     /* content encoding starting point: position */
         encoding_len;     /* content encoding length */
 
+    char *tmp;
     magic_rsl *frag;      /* list-traversal pointer */
     rsl_states state;
 
@@ -722,7 +723,7 @@ static int magic_rsl_to_request(request_rec *r)
                 else {
                     /* should not be possible */
                     /* abandon malfunctioning module */
-                    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+                    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01509)
                                 MODNAME ": bad state %d (ws)", state);
                     return DECLINED;
                 }
@@ -766,7 +767,7 @@ static int magic_rsl_to_request(request_rec *r)
                 else {
                     /* should not be possible */
                     /* abandon malfunctioning module */
-                    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+                    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01510)
                                 MODNAME ": bad state %d (ns)", state);
                     return DECLINED;
                 }
@@ -784,17 +785,13 @@ static int magic_rsl_to_request(request_rec *r)
     }
 
     /* save the info in the request record */
-    if (state == rsl_subtype || state == rsl_encoding ||
-        state == rsl_encoding) {
-        char *tmp;
-        tmp = rsl_strdup(r, type_frag, type_pos, type_len);
-        /* XXX: this could be done at config time I'm sure... but I'm
-         * confused by all this magic_rsl stuff. -djg */
-        ap_content_type_tolower(tmp);
-        ap_set_content_type(r, tmp);
-    }
+    tmp = rsl_strdup(r, type_frag, type_pos, type_len);
+    /* XXX: this could be done at config time I'm sure... but I'm
+     * confused by all this magic_rsl stuff. -djg */
+    ap_content_type_tolower(tmp);
+    ap_set_content_type(r, tmp);
+
     if (state == rsl_encoding) {
-        char *tmp;
         tmp = rsl_strdup(r, encoding_frag,
                                          encoding_pos, encoding_len);
         /* XXX: this could be done at config time I'm sure... but I'm
@@ -806,7 +803,7 @@ static int magic_rsl_to_request(request_rec *r)
     /* detect memory allocation or other errors */
     if (!r->content_type ||
         (state == rsl_encoding && !r->content_encoding)) {
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01511)
                       MODNAME ": unexpected state %d; could be caused by bad "
                       "data in magic file",
                       state);
@@ -845,7 +842,7 @@ static int magic_process(request_rec *r)
 
     if (apr_file_open(&fd, r->filename, APR_READ, APR_OS_DEFAULT, r->pool) != APR_SUCCESS) {
         /* We can't open it, but we were able to stat it. */
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01512)
                     MODNAME ": can't read `%s'", r->filename);
         /* let some other handler decide what the problem is */
         return DECLINED;
@@ -856,7 +853,7 @@ static int magic_process(request_rec *r)
      */
     nbytes = sizeof(buf) - 1;
     if ((result = apr_file_read(fd, (char *) buf, &nbytes)) != APR_SUCCESS) {
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, result, r,
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, result, r, APLOGNO(01513)
                     MODNAME ": read failed: %s", r->filename);
         return HTTP_INTERNAL_SERVER_ERROR;
     }
@@ -885,10 +882,10 @@ static int tryit(request_rec *r, unsigned char *buf, apr_size_t nb,
     /*
      * Try compression stuff
      */
-        if (checkzmagic == 1) {
-                        if (zmagic(r, buf, nb) == 1)
-                        return OK;
-        }
+    if (checkzmagic == 1) {
+        if (zmagic(r, buf, nb) == 1)
+            return OK;
+    }
 
     /*
      * try tests in /etc/magic (or surrogate magic file)
@@ -930,13 +927,13 @@ static int apprentice(server_rec *s, apr_pool_t *p)
     const char *fname = ap_server_root_relative(p, conf->magicfile);
 
     if (!fname) {
-        ap_log_error(APLOG_MARK, APLOG_ERR, APR_EBADPATH, s,
+        ap_log_error(APLOG_MARK, APLOG_ERR, APR_EBADPATH, s, APLOGNO(01514)
                      MODNAME ": Invalid magic file path %s", conf->magicfile);
         return -1;
     }
     if ((result = apr_file_open(&f, fname, APR_READ | APR_BUFFERED,
-                                APR_OS_DEFAULT, p) != APR_SUCCESS)) {
-        ap_log_error(APLOG_MARK, APLOG_ERR, result, s,
+                                APR_OS_DEFAULT, p)) != APR_SUCCESS) {
+        ap_log_error(APLOG_MARK, APLOG_ERR, result, s, APLOGNO(01515)
                      MODNAME ": can't read magic file %s", fname);
         return -1;
     }
@@ -986,28 +983,28 @@ static int apprentice(server_rec *s, apr_pool_t *p)
     (void) apr_file_close(f);
 
 #if MIME_MAGIC_DEBUG
-    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, APLOGNO(01516)
                 MODNAME ": apprentice conf=%x file=%s m=%s m->next=%s last=%s",
                 conf,
                 conf->magicfile ? conf->magicfile : "NULL",
                 conf->magic ? "set" : "NULL",
                 (conf->magic && conf->magic->next) ? "set" : "NULL",
                 conf->last ? "set" : "NULL");
-    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, APLOGNO(01517)
                 MODNAME ": apprentice read %d lines, %d rules, %d errors",
                 lineno, rule, errs);
 #endif
 
 #if MIME_MAGIC_DEBUG
     prevm = 0;
-    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, APLOGNO(01518)
                 MODNAME ": apprentice test");
     for (m = conf->magic; m; m = m->next) {
         if (apr_isprint((((unsigned long) m) >> 24) & 255) &&
             apr_isprint((((unsigned long) m) >> 16) & 255) &&
             apr_isprint((((unsigned long) m) >> 8) & 255) &&
             apr_isprint(((unsigned long) m) & 255)) {
-            ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+            ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, APLOGNO(01519)
                         MODNAME ": apprentice: POINTER CLOBBERED! "
                         "m=\"%c%c%c%c\" line=%d",
                         (((unsigned long) m) >> 24) & 255,
@@ -1054,7 +1051,7 @@ static unsigned long signextend(server_rec *s, struct magic *m, unsigned long v)
         case STRING:
             break;
         default:
-            ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
+            ap_log_error(APLOG_MARK, APLOG_ERR, 0, s, APLOGNO(01520)
                         MODNAME ": can't happen: m->type=%d", m->type);
             return -1;
         }
@@ -1102,7 +1099,7 @@ static int parse(server_rec *serv, apr_pool_t *p, char *l, int lineno)
     /* get offset, then skip over it */
     m->offset = (int) strtol(l, &t, 0);
     if (l == t) {
-        ap_log_error(APLOG_MARK, APLOG_ERR, 0, serv,
+        ap_log_error(APLOG_MARK, APLOG_ERR, 0, serv, APLOGNO(01521)
                     MODNAME ": offset %s invalid", l);
     }
     l = t;
@@ -1125,7 +1122,7 @@ static int parse(server_rec *serv, apr_pool_t *p, char *l, int lineno)
                 m->in.type = BYTE;
                 break;
             default:
-                ap_log_error(APLOG_MARK, APLOG_ERR, 0, serv,
+                ap_log_error(APLOG_MARK, APLOG_ERR, 0, serv, APLOGNO(01522)
                         MODNAME ": indirect offset type %c invalid", *l);
                 break;
             }
@@ -1142,7 +1139,7 @@ static int parse(server_rec *serv, apr_pool_t *p, char *l, int lineno)
         else
             t = l;
         if (*t++ != ')') {
-            ap_log_error(APLOG_MARK, APLOG_ERR, 0, serv,
+            ap_log_error(APLOG_MARK, APLOG_ERR, 0, serv, APLOGNO(01523)
                         MODNAME ": missing ')' in indirect offset");
         }
         l = t;
@@ -1216,7 +1213,7 @@ static int parse(server_rec *serv, apr_pool_t *p, char *l, int lineno)
         l += NLEDATE;
     }
     else {
-        ap_log_error(APLOG_MARK, APLOG_ERR, 0, serv,
+        ap_log_error(APLOG_MARK, APLOG_ERR, 0, serv, APLOGNO(01524)
                     MODNAME ": type %s invalid", l);
         return -1;
     }
@@ -1275,11 +1272,10 @@ static int parse(server_rec *serv, apr_pool_t *p, char *l, int lineno)
     }
     else
         m->nospflag = 0;
-    strncpy(m->desc, l, sizeof(m->desc) - 1);
-    m->desc[sizeof(m->desc) - 1] = '\0';
+    apr_cpystrn(m->desc, l, sizeof(m->desc));
 
 #if MIME_MAGIC_DEBUG
-    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, serv,
+    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, serv, APLOGNO(01525)
                 MODNAME ": parse line=%d m=%x next=%x cont=%d desc=%s",
                 lineno, m, m->next, m->cont_level, m->desc);
 #endif /* MIME_MAGIC_DEBUG */
@@ -1322,7 +1318,7 @@ static char *getstr(server_rec *serv, register char *s, register char *p,
         if (apr_isspace(c))
             break;
         if (p >= pmax) {
-            ap_log_error(APLOG_MARK, APLOG_ERR, 0, serv,
+            ap_log_error(APLOG_MARK, APLOG_ERR, 0, serv, APLOGNO(01526)
                         MODNAME ": string too long: %s", origs);
             break;
         }
@@ -1468,7 +1464,7 @@ static int fsmagic(request_rec *r, const char *fn)
         /* We used stat(), the only possible reason for this is that the
          * symlink is broken.
          */
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01527)
                     MODNAME ": broken symlink (%s)", fn);
         return HTTP_INTERNAL_SERVER_ERROR;
     case APR_SOCK:
@@ -1477,7 +1473,7 @@ static int fsmagic(request_rec *r, const char *fn)
     case APR_REG:
         break;
     default:
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01528)
                       MODNAME ": invalid file type %d.", r->finfo.filetype);
         return HTTP_INTERNAL_SERVER_ERROR;
     }
@@ -1544,7 +1540,7 @@ static int match(request_rec *r, unsigned char *s, apr_size_t nbytes)
     struct magic *m;
 
 #if MIME_MAGIC_DEBUG
-    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01529)
                 MODNAME ": match conf=%x file=%s m=%s m->next=%s last=%s",
                 conf,
                 conf->magicfile ? conf->magicfile : "NULL",
@@ -1559,7 +1555,7 @@ static int match(request_rec *r, unsigned char *s, apr_size_t nbytes)
             apr_isprint((((unsigned long) m) >> 16) & 255) &&
             apr_isprint((((unsigned long) m) >> 8) & 255) &&
             apr_isprint(((unsigned long) m) & 255)) {
-            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01530)
                         MODNAME ": match: POINTER CLOBBERED! "
                         "m=\"%c%c%c%c\"",
                         (((unsigned long) m) >> 24) & 255,
@@ -1574,7 +1570,7 @@ static int match(request_rec *r, unsigned char *s, apr_size_t nbytes)
     for (m = conf->magic; m; m = m->next) {
 #if MIME_MAGIC_DEBUG
         rule_counter++;
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01531)
                     MODNAME ": line=%d desc=%s", m->lineno, m->desc);
 #endif
 
@@ -1594,7 +1590,7 @@ static int match(request_rec *r, unsigned char *s, apr_size_t nbytes)
             while (m_cont && (m_cont->cont_level != 0)) {
 #if MIME_MAGIC_DEBUG
                 rule_counter++;
-                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01532)
                         MODNAME ": line=%d mc=%x mc->next=%x cont=%d desc=%s",
                             m_cont->lineno, m_cont,
                             m_cont->next, m_cont->cont_level,
@@ -1613,7 +1609,7 @@ static int match(request_rec *r, unsigned char *s, apr_size_t nbytes)
         /* if we get here, the main entry rule was a match */
         /* this will be the last run through the loop */
 #if MIME_MAGIC_DEBUG
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01533)
                     MODNAME ": rule matched, line=%d type=%d %s",
                     m->lineno, m->type,
                     (m->type == STRING) ? m->value.s : "");
@@ -1637,7 +1633,7 @@ static int match(request_rec *r, unsigned char *s, apr_size_t nbytes)
         m = m->next;
         while (m && (m->cont_level != 0)) {
 #if MIME_MAGIC_DEBUG
-            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01534)
                         MODNAME ": match line=%d cont=%d type=%d %s",
                         m->lineno, m->cont_level, m->type,
                         (m->type == STRING) ? m->value.s : "");
@@ -1681,13 +1677,13 @@ static int match(request_rec *r, unsigned char *s, apr_size_t nbytes)
             m = m->next;
         }
 #if MIME_MAGIC_DEBUG
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01535)
                     MODNAME ": matched after %d rules", rule_counter);
 #endif
         return 1;  /* all through */
     }
 #if MIME_MAGIC_DEBUG
-    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01536)
                 MODNAME ": failed after %d rules", rule_counter);
 #endif
     return 0;  /* no match at all */
@@ -1733,7 +1729,7 @@ static void mprint(request_rec *r, union VALUETYPE *p, struct magic *m)
         (void) magic_rsl_printf(r, m->desc, pp);
         return;
     default:
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01537)
                     MODNAME ": invalid m->type (%d) in mprint().",
                     m->type);
         return;
@@ -1779,7 +1775,7 @@ static int mconvert(request_rec *r, union VALUETYPE *p, struct magic *m)
             ((p->hl[3] << 24) | (p->hl[2] << 16) | (p->hl[1] << 8) | (p->hl[0]));
         return 1;
     default:
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01538)
                     MODNAME ": invalid type %d in mconvert().", m->type);
         return 0;
     }
@@ -1831,7 +1827,7 @@ static int mcheck(request_rec *r, union VALUETYPE *p, struct magic *m)
     int matched;
 
     if ((m->value.s[0] == 'x') && (m->value.s[1] == '\0')) {
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01539)
                     MODNAME ": BOINK");
         return 1;
     }
@@ -1876,7 +1872,7 @@ static int mcheck(request_rec *r, union VALUETYPE *p, struct magic *m)
         break;
     default:
         /*  bogosity, pretend that it just wasn't a match */
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01540)
                     MODNAME ": invalid type %d in mcheck().", m->type);
         return 0;
     }
@@ -1886,7 +1882,7 @@ static int mcheck(request_rec *r, union VALUETYPE *p, struct magic *m)
     switch (m->reln) {
     case 'x':
 #if MIME_MAGIC_DEBUG
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01541)
                     "%lu == *any* = 1", v);
 #endif
         matched = 1;
@@ -1895,7 +1891,7 @@ static int mcheck(request_rec *r, union VALUETYPE *p, struct magic *m)
     case '!':
         matched = v != l;
 #if MIME_MAGIC_DEBUG
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01542)
                     "%lu != %lu = %d", v, l, matched);
 #endif
         break;
@@ -1903,7 +1899,7 @@ static int mcheck(request_rec *r, union VALUETYPE *p, struct magic *m)
     case '=':
         matched = v == l;
 #if MIME_MAGIC_DEBUG
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01543)
                     "%lu == %lu = %d", v, l, matched);
 #endif
         break;
@@ -1912,14 +1908,14 @@ static int mcheck(request_rec *r, union VALUETYPE *p, struct magic *m)
         if (m->flag & UNSIGNED) {
             matched = v > l;
 #if MIME_MAGIC_DEBUG
-            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01544)
                         "%lu > %lu = %d", v, l, matched);
 #endif
         }
         else {
             matched = (long) v > (long) l;
 #if MIME_MAGIC_DEBUG
-            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01545)
                         "%ld > %ld = %d", v, l, matched);
 #endif
         }
@@ -1929,14 +1925,14 @@ static int mcheck(request_rec *r, union VALUETYPE *p, struct magic *m)
         if (m->flag & UNSIGNED) {
             matched = v < l;
 #if MIME_MAGIC_DEBUG
-            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01546)
                         "%lu < %lu = %d", v, l, matched);
 #endif
         }
         else {
             matched = (long) v < (long) l;
 #if MIME_MAGIC_DEBUG
-            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01547)
                         "%ld < %ld = %d", v, l, matched);
 #endif
         }
@@ -1945,7 +1941,7 @@ static int mcheck(request_rec *r, union VALUETYPE *p, struct magic *m)
     case '&':
         matched = (v & l) == l;
 #if MIME_MAGIC_DEBUG
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01548)
                     "((%lx & %lx) == %lx) = %d", v, l, l, matched);
 #endif
         break;
@@ -1953,7 +1949,7 @@ static int mcheck(request_rec *r, union VALUETYPE *p, struct magic *m)
     case '^':
         matched = (v & l) != l;
 #if MIME_MAGIC_DEBUG
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01549)
                     "((%lx & %lx) != %lx) = %d", v, l, l, matched);
 #endif
         break;
@@ -1961,7 +1957,7 @@ static int mcheck(request_rec *r, union VALUETYPE *p, struct magic *m)
     default:
         /* bogosity, pretend it didn't match */
         matched = 0;
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01550)
                     MODNAME ": mcheck: can't happen: invalid relation %d.",
                     m->reln);
         break;
@@ -1977,9 +1973,9 @@ static int ascmagic(request_rec *r, unsigned char *buf, apr_size_t nbytes)
 {
     int has_escapes = 0;
     unsigned char *s;
-    char nbuf[HOWMANY + 1];  /* one extra for terminating '\0' */
+    char nbuf[SMALL_HOWMANY + 1];  /* one extra for terminating '\0' */
     char *token;
-    register struct names *p;
+    const struct names *p;
     int small_nbytes;
     char *strtok_state;
 
@@ -2099,13 +2095,15 @@ static int zmagic(request_rec *r, unsigned char *buf, apr_size_t nbytes)
     if (i == ncompr)
         return 0;
 
-    if ((newsize = uncompress(r, i, &newbuf, nbytes)) > 0) {
+    if ((newsize = uncompress(r, i, &newbuf, HOWMANY)) > 0) {
+        /* set encoding type in the request record */
+        r->content_encoding = compr[i].encoding;
+
+        newbuf[newsize-1] = '\0';  /* null-terminate uncompressed data */
+        /* Try to detect the content type of the uncompressed data */
         if (tryit(r, newbuf, newsize, 0) != OK) {
             return 0;
         }
-
-        /* set encoding type in the request record */
-        r->content_encoding = compr[i].encoding;
     }
     return 1;
 }
@@ -2121,7 +2119,6 @@ static int create_uncompress_child(struct uncompress_parms *parm, apr_pool_t *cn
 {
     int rc = 1;
     const char *new_argv[4];
-    const char *const *env;
     request_rec *r = parm->r;
     apr_pool_t *child_context = cntxt;
     apr_procattr_t *procattr;
@@ -2133,15 +2130,14 @@ static int create_uncompress_child(struct uncompress_parms *parm, apr_pool_t *cn
      * Should we create the err pipe, read it, and copy to the log?
      */
 
-    env = (const char *const *)ap_create_environment(child_context, r->subprocess_env);
-
     if ((apr_procattr_create(&procattr, child_context) != APR_SUCCESS) ||
         (apr_procattr_io_set(procattr, APR_FULL_BLOCK,
                            APR_FULL_BLOCK, APR_NO_PIPE)   != APR_SUCCESS) ||
-        (apr_procattr_dir_set(procattr, r->filename)        != APR_SUCCESS) ||
-        (apr_procattr_cmdtype_set(procattr, APR_PROGRAM)    != APR_SUCCESS)) {
+        (apr_procattr_dir_set(procattr,
+                              ap_make_dirstr_parent(r->pool, r->filename)) != APR_SUCCESS) ||
+        (apr_procattr_cmdtype_set(procattr, APR_PROGRAM_PATH) != APR_SUCCESS)) {
         /* Something bad happened, tell the world. */
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, APR_ENOPROC, r,
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, APR_ENOPROC, r, APLOGNO(01551)
                "couldn't setup child process: %s", r->filename);
     }
     else {
@@ -2152,11 +2148,11 @@ static int create_uncompress_child(struct uncompress_parms *parm, apr_pool_t *cn
 
         procnew = apr_pcalloc(child_context, sizeof(*procnew));
         rc = apr_proc_create(procnew, compr[parm->method].argv[0],
-                               new_argv, env, procattr, child_context);
+                               new_argv, NULL, procattr, child_context);
 
         if (rc != APR_SUCCESS) {
             /* Bad things happened. Everyone should have cleaned up. */
-            ap_log_rerror(APLOG_MARK, APLOG_ERR, APR_ENOPROC, r,
+            ap_log_rerror(APLOG_MARK, APLOG_ERR, APR_ENOPROC, r, APLOGNO(01552)
                           MODNAME ": could not execute `%s'.",
                           compr[parm->method].argv[0]);
         }
@@ -2188,7 +2184,7 @@ static int uncompress(request_rec *r, int method,
         return -1;
 
     if ((rv = create_uncompress_child(&parm, sub_context, &pipe_out)) != APR_SUCCESS) {
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r, APLOGNO(01553)
                     MODNAME ": couldn't spawn uncompress process: %s", r->uri);
         return -1;
     }
@@ -2197,7 +2193,7 @@ static int uncompress(request_rec *r, int method,
     rv = apr_file_read(pipe_out, *newch, &n);
     if (n == 0) {
         apr_pool_destroy(sub_context);
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r, APLOGNO(01554)
             MODNAME ": read failed from uncompress of %s", r->filename);
         return -1;
     }
@@ -2304,7 +2300,7 @@ static int revision_suffix(request_rec *r)
     request_rec *sub;
 
 #if MIME_MAGIC_DEBUG
-    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01555)
                 MODNAME ": revision_suffix checking %s", r->filename);
 #endif /* MIME_MAGIC_DEBUG */
 
@@ -2323,7 +2319,7 @@ static int revision_suffix(request_rec *r)
     result = 0;
     sub_filename = apr_pstrndup(r->pool, r->filename, suffix_pos);
 #if MIME_MAGIC_DEBUG
-    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01556)
                 MODNAME ": subrequest lookup for %s", sub_filename);
 #endif /* MIME_MAGIC_DEBUG */
     sub = ap_sub_req_lookup_file(sub_filename, r, NULL);
@@ -2332,7 +2328,7 @@ static int revision_suffix(request_rec *r)
     if (sub->content_type) {
         ap_set_content_type(r, apr_pstrdup(r->pool, sub->content_type));
 #if MIME_MAGIC_DEBUG
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01557)
                     MODNAME ": subrequest %s got %s",
                     sub_filename, r->content_type);
 #endif /* MIME_MAGIC_DEBUG */
@@ -2383,14 +2379,14 @@ static int magic_init(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server
                 return OK;
 #if MIME_MAGIC_DEBUG
             prevm = 0;
-            ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+            ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, APLOGNO(01558)
                         MODNAME ": magic_init 1 test");
             for (m = conf->magic; m; m = m->next) {
                 if (apr_isprint((((unsigned long) m) >> 24) & 255) &&
                     apr_isprint((((unsigned long) m) >> 16) & 255) &&
                     apr_isprint((((unsigned long) m) >> 8) & 255) &&
                     apr_isprint(((unsigned long) m) & 255)) {
-                    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+                    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, APLOGNO(01559)
                                 MODNAME ": magic_init 1: POINTER CLOBBERED! "
                                 "m=\"%c%c%c%c\" line=%d",
                                 (((unsigned long) m) >> 24) & 255,
@@ -2418,7 +2414,7 @@ static int magic_find_ct(request_rec *r)
     magic_server_config_rec *conf;
 
     /* the file has to exist */
-    if (r->finfo.filetype == 0 || !r->filename) {
+    if (r->finfo.filetype == APR_NOFILE || !r->filename) {
         return DECLINED;
     }
 
@@ -2463,7 +2459,7 @@ static void register_hooks(apr_pool_t *p)
  * Apache API module interface
  */
 
-module AP_MODULE_DECLARE_DATA mime_magic_module =
+AP_DECLARE_MODULE(mime_magic) =
 {
     STANDARD20_MODULE_STUFF,
     NULL,                      /* dir config creator */
@@ -2473,5 +2469,3 @@ module AP_MODULE_DECLARE_DATA mime_magic_module =
     mime_magic_cmds,           /* command apr_table_t */
     register_hooks              /* register hooks */
 };
-
-
